"""Processing operations that have much to do with coordinates."""
from __future__ import annotations

from collections.abc import Sequence

from typing_extensions import Self

from dascore.constants import PatchType
from dascore.core.coords import BaseCoord
from dascore.exceptions import CoordError
from dascore.utils.misc import get_parent_code_name, iterate
from dascore.utils.patch import patch_function


@patch_function()
def snap_coords(patch: PatchType, *coords, reverse=False):
    """
    Snap coordinates to evenly spaced samples.

    This ensures all of the specified coordinates are evenly spaced and
    monotonic. First, the patch is sorted along specified coordinates,
    then coordinates are assumed evenly-sampled from their min to max value.
    Doing this can introduce some error since the coordinate labels are moved
    and the data (apart from the sorting) are left unchanged.
    Consider using [interpolate](`dascore.Patch.interpolate`) for a more expensive
    but more accurate linear interpolation.

    Parameters
    ----------
    *coords
        Used to specify the dimension names to convert to CoordRanges. If not
        specified convert all dimensional coordinates.
    reverse
        If True, reverse the sorting of the coordinates.

    Examples
    --------
    >>> import dascore as dc
    >>> # get an example patch which has unevenly sampled coords time, distance
    >>> patch = dc.get_example_patch("wacky_dim_coords_patch")
    >>> # snap time dimension
    >>> time_snap = patch.snap_coords("time")
    >>> # snap the distance dimension
    >>> dist_snap = patch.snap_coords("distance")
    """
    cman, data = patch.coords.snap(*coords, array=patch.data, reverse=reverse)
    return patch.new(data=data, coords=cman)


@patch_function()
def sort_coords(patch: PatchType, *coords, reverse=False):
    """
    Sort one or more coordinates.

    Sorts the specified coordinates in the patch. An error will be raised
    if the coordinates have overlapping dimensions since it may not be
    possible to sort each. An error is also raised in any of the coordinates
    are multidimensional.

    Parameters
    ----------
    *coords
        Used to specify the coordinates to sort.
    reverse
        If True, sort in descending order, else ascending.

    Examples
    --------
    >>> import dascore as dc
    >>> # get an example patch which has unevenly sampled coords time, distance
    >>> patch = dc.get_example_patch("wacky_dim_coords_patch")
    >>> # sort time coordinate (dimension) in ascending order
    >>> time_snap = patch.sort_coords("time")
    >>> assert time_snap.coords.coord_map['time'].sorted
    >>> # sort distance coordinate (dimension) in descending order
    >>> dist_snap = patch.sort_coords("distance", reverse=True)
    >>> assert dist_snap.coords.coord_map['distance'].reverse_sorted
    """
    cman, data = patch.coords.sort(*coords, array=patch.data, reverse=reverse)
    return patch.new(data=data, coords=cman)


def get_coord(
    self: PatchType,
    name: str,
    require_sorted: bool = False,
    require_evenly_sampled: bool = False,
) -> BaseCoord:
    """
    Get a managed coordinate, raising if it doesn't meet requirements.

    Parameters
    ----------
    name
        Name of the coordinate to fetch.
    require_sorted
        If True, require the coordinate to be sorted or raise Error.
    require_evenly_sampled
        If True, require the coordinate to be evenly sampled or raise Error.
    """
    coord = self.coords.coord_map[name]
    if require_evenly_sampled and coord.step is None:
        extra = f"as required by {get_parent_code_name()}"  # adds caller name
        msg = f"Coordinate {name} is not evenly sampled {extra}"
        raise CoordError(msg)
    if require_sorted and not coord.sorted or coord.reverse_sorted:
        extra = f"as required by {get_parent_code_name()}"  # adds caller name
        msg = f"Coordinate {name} is not sorted {extra}"
        raise CoordError(msg)
    return coord


def assert_has_coords(self: PatchType, coord_names: Sequence[str] | str) -> Self:
    """Raise an error if patch doesn't have required coordinates."""
    required_coords = set(iterate(coord_names))
    current_coords = set(self.coords.coord_map)
    if missing := required_coords - current_coords:
        msg = f"Patch does not have required coordinate(s): {missing}"
        raise CoordError(msg)
    return self


@patch_function()
def rename_coords(self: PatchType, **kwargs) -> PatchType:
    """
    Rename coordinate of Patch.

    Parameters
    ----------
    **kwargs
        The mapping from old names to new names

    Examples
    --------
    >>> import dascore as dc
    >>> pa = dc.get_example_patch()
    >>> # rename dim "distance" to "fragrance"
    >>> pa2 = pa.rename_coords(distance='fragrance')
    >>> assert 'fragrance' in pa2.dims
    """
    new_coord = self.coords.rename_coord(**kwargs)
    attrs = self.attrs.rename_dimension(**kwargs)
    return self.new(coords=new_coord, dims=new_coord.dims, attrs=attrs)


@patch_function()
def update_coords(self: PatchType, **kwargs) -> PatchType:
    """
    Update the coordiantes of a patch.

    Will either add new coordinates, or update existing ones.

    Parameters
    ----------
    **kwargs
        The mapping from old names to new names

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> pa = dc.get_example_patch()
    >>> # Add 1 to all distance coords
    >>> new_dist = pa.coords['distance'] + 1
    >>> pa2 = pa.update_coords(distance=new_dist)
    >>> assert np.all(pa2.coords['distance'] == (pa.coords['distance'] + 1))
    """
    new_coord = self.coords.update_coords(**kwargs)
    return self.new(coords=new_coord, dims=new_coord.dims)



"""Module for testing units."""
from __future__ import annotations

import numpy as np
import pytest

import dascore as dc
from dascore.exceptions import UnitError
from dascore.units import (
    Quantity,
    assert_dtype_compatible_with_units,
    convert_units,
    get_factor_and_unit,
    get_filter_units,
    get_quantity,
    get_quantity_str,
    get_unit,
    invert_quantity,
)


class TestUnitInit:
    """Ensure units can be initialized."""

    def test_time(self):
        """Tests for time units."""
        sec = dc.get_unit("s")
        assert str(sec.dimensionality) == "[time]"

    def test_invert(self):
        """Ensure a unit can be inverted."""
        unit = dc.get_unit("s")
        inverted = invert_quantity(unit)
        reverted = invert_quantity(inverted)
        assert unit == reverted
        assert invert_quantity(None) is None


class TestGetQuantStr:
    """Ensure units can be validated."""

    valid = ("m", "s", "Hz", "1/s", "1/m", "feet", "furlongs", "km", "fortnight")
    invalid = ("bob", "gerbil inches", "farsee")

    @pytest.mark.parametrize("in_str", valid)
    def test_validate_units_good_input(self, in_str):
        """Ensure units can be validated from various inputs."""
        assert get_quantity_str(in_str)

    @pytest.mark.parametrize("in_str", invalid)
    def test_validate_units_bad_input(self, in_str):
        """Ensure units can be validated from various inputs."""
        with pytest.raises(UnitError):
            get_quantity_str(in_str)

    def test_none(self):
        """Ensure none and empty str also works."""
        assert get_quantity_str(None) is None
        assert get_quantity_str("") is None

    def test_quantity(self):
        """Ensure a quantity works."""
        # with no magnitude the string should be simple units
        quant = get_quantity("m/s")
        out = get_quantity_str(quant)
        assert out == "m / s"
        # with magnitude it should be included.
        quant = get_quantity("10 m /s")
        out = get_quantity_str(quant)
        assert "10.0" in out


class TestUnitAndFactor:
    """tests for returning units and scaling factor."""

    def test_quantx_units(self):
        """Tests for the quantx unit str."""
        mag, ustr = get_factor_and_unit("rad * 2pi/2^16")
        assert ustr == "rad * π"
        assert np.isclose(mag, (2 / (2**16)))

    def test_simplify_units(self):
        """Test for reducing units."""
        mag, ustr = get_factor_and_unit("rad * (km/m)", simplify=True)
        assert get_unit(ustr) == get_unit("radian")
        assert np.isclose(mag, 1000)

    def test_none(self):
        """Ensure none returns a None and string."""
        factor, unit = get_factor_and_unit(None)
        assert factor == 1
        assert unit is None


class TestGetQuantity:
    """Tests for getting a quantity."""

    def test_quantity_identity(self):
        """Get quantity should always return the same quantity."""
        quant1 = get_quantity("1/s")
        quant2 = get_quantity("1 Hz")
        assert quant1 == get_quantity(quant1)
        assert quant1 is get_quantity(quant1)
        assert quant2 == get_quantity(quant2)
        assert quant2 is get_quantity(quant2)

    def test_get_temp(self):
        """Get quantity should work with temperatures."""
        quant1 = get_quantity("degC")
        assert "°C" in str(quant1)


class TestConvenientImport:
    """Tests for conveniently importing units for dascore.units."""

    def test_import_common(self):
        """Ensure common units are importable."""
        from dascore.units import Hz, ft, km, m, miles

        assert m == get_unit("m")
        assert ft == get_unit("ft")
        assert miles == get_unit("miles")
        assert km == get_unit("km")
        assert Hz == get_unit("Hz")

    def test_bad_import_error_msg(self):
        """An import error should be raised if the unit isn't valid."""
        with pytest.raises(ImportError):
            from dascore.utils import bob  # noqa


class TestGetFilterUnits:
    """Tests for getting units that can be used for filtering."""

    def test_no_units(self):
        """Tests for when no units are specified."""
        assert get_filter_units(1, 10, "m") == (1.0, 10.0)
        assert get_filter_units(None, 10, "s") == (None, 10.0)
        assert get_filter_units(1, None, "s") == (1.0, None)

    def test_filter_units(self):
        """Tests for when filter units are already those selected."""
        Hz = get_unit("Hz")
        s = get_unit("s")
        assert get_filter_units(1.0 * Hz, 10.0 * Hz, s) == (1.0, 10.0)
        assert get_filter_units(None, 10.0 * Hz, s) == (None, 10.0)
        assert get_filter_units(1.0 * Hz, 10.0 * Hz, s) == (1.0, 10.0)

    def test_same_units(self):
        """Tests for when filter units are already those selected."""
        s = get_unit("s")
        assert get_filter_units(1.0 * s, 10.0 * s, s) == (0.1, 1.0)
        assert get_filter_units(None, 10.0 * s, s) == (None, 0.1)
        assert get_filter_units(10.0 * s, None, s) == (0.1, None)

    def test_different_units_raises(self):
        """The units must be the same or it should raise."""
        s, Hz = get_unit("s"), get_unit("Hz")

        with pytest.raises(UnitError):
            get_filter_units(1.0, 10.0 * s, s)

        with pytest.raises(UnitError):
            get_filter_units(1.0 * s, 10.0 * Hz, s)

    def test_incompatible_units_raise(self):
        """The units must be the same or it should raise."""
        s, m = get_unit("s"), get_unit("m")
        match = "Cannot convert from"
        with pytest.raises(UnitError, match=match):
            get_filter_units(1.0 * s, 10.0 * s, m)

        with pytest.raises(UnitError, match=match):
            get_filter_units(1.0 * m, 10.0 * m, s)


class TestDTypeCompatible:
    """Ensure dtype compatibility check works."""

    quants = ("degC", "m/s", get_quantity("kg"))
    non_dt_dtypes = (np.float_, np.int_, np.float32)

    def test_non_datetime(self):
        """Any non-datetime should be compatible."""
        for quant in self.quants:
            for dtype in self.non_dt_dtypes:
                out = assert_dtype_compatible_with_units(dtype, quant)
                assert isinstance(out, Quantity)

    def test_bad_dim_raises(self):
        """Ensure a bad dimension of quantity raises."""
        for quant in self.quants:
            with pytest.raises(UnitError):
                assert_dtype_compatible_with_units(np.datetime64, quant)
            with pytest.raises(UnitError):
                assert_dtype_compatible_with_units(np.timedelta64, quant)

    def test_non_s_raises(self):
        """Only 's' should work, no other increment of time."""
        match = "only allowable units are s"
        with pytest.raises(UnitError, match=match):
            assert_dtype_compatible_with_units(np.datetime64, "ms")

    def test_s_works(self):
        """Seconds should work fine."""
        out = assert_dtype_compatible_with_units(np.datetime64, "s")
        assert out == get_quantity("s")


class TestConvertUnits:
    """Test suite for converting units."""

    def test_simple(self):
        """Simple units to simple units."""
        out = convert_units(1, "m", "ft")
        assert np.isclose(out, 0.3048)

    def test_temperature(self):
        """Ensure temperature can be converted."""
        out = convert_units(1, "m", "ft")
        assert np.isclose(out, 0.3048)

    def test_convert_offset_units(self):
        """Test simple offset units."""
        array = np.arange(10)
        f_array = array * (9 / 5) + 32.0
        out = convert_units(array, from_units="degC", to_units="degF")
        assert np.allclose(f_array, out)

    def test_convert_offset_units_with_mag(self):
        """Ensure units can be converted/set for offset units when non-1 magnitudes."""
        # One non-1 quantity
        array = np.arange(10)
        f_array = 2 * array * (9 / 5) + 32.0
        out = convert_units(array, from_units="2*degC", to_units="degF")
        assert np.allclose(f_array, out)

    def test_convert_offset_units_multiple_mags(self):
        """Ensure if both units have non-1 offsets conversion still works."""
        # Multiple non-1 quants
        array = np.arange(10)
        f_array = (array * (18 / 5) + 32.0) / 2
        out = convert_units(array, from_units="2*degC", to_units="2*degF")
        assert np.allclose(f_array, out)
        # non equal quants
        f_array = (array * (9 * 2.5 / 5) + 32.0) / 6
        out = convert_units(array, from_units="2.5*degC", to_units="6*degF")
        assert np.allclose(f_array, out)



"""Utilities for basic IO tasks."""
from __future__ import annotations

import abc
import typing
from contextlib import suppress
from functools import cache, singledispatch
from inspect import isfunction, ismethod
from pathlib import Path
from typing import Any, Protocol, get_type_hints

import tables

from dascore.utils.misc import cached_method, register_func

HANDLE_FUNCTIONS = {}


RequiredType = typing.TypeVar("RequiredType")


@typing.runtime_checkable
class BinaryWriter(Protocol):
    """Dummy class for streams which write binary."""

    closed: bool

    @abc.abstractmethod
    def read(self, *args, **kwargs):
        """Read resource contents."""

    @abc.abstractmethod
    def seek(self, *args, **kwargs):
        """Seek to specific place in file."""

    @abc.abstractmethod
    def close(self, *args, **kwargs):
        """Close resource."""


@typing.runtime_checkable
class BinaryReader(Protocol):
    """Dummy class for streams which read binary."""

    closed: bool

    @abc.abstractmethod
    def write(self, *args, **kwargs):
        """Write resource contents."""

    @abc.abstractmethod
    def seek(self, *args, **kwargs):
        """Seek to specific place in file."""

    @abc.abstractmethod
    def close(self, *args, **kwargs):
        """Close resource."""


@typing.runtime_checkable
class HDF5Writer(Protocol):
    """An HDF 5 file open in write mode."""

    root: Any
    params: Any
    title: str
    filters: Any
    isopen: bool
    filename: str


@typing.runtime_checkable
class HDF5Reader(Protocol):
    """An HDF 5 file open in read mode."""

    root: Any
    params: Any
    title: str
    filters: Any
    isopen: bool
    filename: str


@cache
def _get_required_type(required_type, arg_name=None):
    """Get the type hint for the first argument."""
    if required_type not in HANDLE_FUNCTIONS:
        # here we try to get the type from the function type hints
        # but we need to skip things that aren't functions
        is_func_y = isfunction(required_type) or ismethod(required_type)
        if not is_func_y or not (hints := get_type_hints(required_type)):
            return required_type
        arg_name = arg_name if arg_name is not None else next(iter(hints))
        return hints.get(arg_name)
    return required_type


def get_handle_from_resource(uri, required_type):
    """
    Get a handle for a file of preferred type.

    return uri if required type is not specified.
    """
    with suppress(TypeError):
        if isinstance(uri, required_type):
            return uri
    if (func := HANDLE_FUNCTIONS.get(required_type)) is None:
        return uri
    return func(uri)


@register_func(HANDLE_FUNCTIONS, key=Path)
def _get_path(uri):
    """Get a path from uri."""
    return Path(uri)


@register_func(HANDLE_FUNCTIONS, key=str)
def _get_str(uri):
    """Get a string from uri."""
    return str(uri)


@register_func(HANDLE_FUNCTIONS, key=BinaryReader)
@singledispatch
def _get_buffered_reader(uri):
    """Get a buffered reader from unique resource."""
    raise NotImplementedError(f"not implemented for type {type(uri)}")


@_get_buffered_reader.register(str)
@_get_buffered_reader.register(Path)
def _uri_to_buffered_reader(path_like):
    """Get a read buffer from as string/path."""
    fi = open(path_like, "rb")
    return fi


@register_func(HANDLE_FUNCTIONS, key=BinaryWriter)
@singledispatch
def _get_buffered_writer(uri):
    """Get a buffered reader from unique resource."""
    raise NotImplementedError(f"not implemented for type {type(uri)}")


@_get_buffered_writer.register(str)
@_get_buffered_writer.register(Path)
def _uri_to_buffered_writer(path_like):
    """Get a write buffer from as string/path."""
    fi = open(path_like, "wb")
    return fi


@register_func(HANDLE_FUNCTIONS, key=HDF5Reader)
@singledispatch
def _get_hdf5_reader(uri):
    """Get a buffered reader from unique resource."""
    raise NotImplementedError(f"not implemented for type {type(uri)}")


@_get_hdf5_reader.register(str)
@_get_hdf5_reader.register(Path)
def _uri_to_hdf_reader(path_like):
    """Get a read buffer from as string/path."""
    fi = tables.open_file(path_like, "r")
    return fi


@register_func(HANDLE_FUNCTIONS, key=HDF5Writer)
@singledispatch
def _get_hdf5_writer(uri):
    """Get a buffered reader from unique resource."""
    raise NotImplementedError(f"not implemented for type {type(uri)}")


@_get_hdf5_writer.register(str)
@_get_hdf5_writer.register(Path)
def _uri_to_hdf5_writer(path_like):
    """Get a write buffer from as string/path."""
    # ensure the file structure exists
    Path(path_like).parent.mkdir(exist_ok=True, parents=True)
    fi = tables.open_file(path_like, "a")
    return fi


class IOResourceManager:
    """A class for managing opening/closing files."""

    def __init__(self, source: Any):
        self._source = source
        self._cache = {}

    @property
    @cached_method
    def source(self):
        """Get the source of the IO manager."""
        source = self._source
        # this handles IO managers derived from other IO managers;
        # effectively, we need to go back to the original, non-io manager source
        while isinstance(source, self.__class__):
            source = source.source
        return source

    def get_resource(self, required_type: RequiredType) -> RequiredType:
        """Get the requested resource from."""
        # no required type, just return source of manager.
        if required_type is None:
            return self.source
        # this is so the context managers can be nested and the child
        # context manager only calls to the parent. Then, the resources
        # get closed only after the original exists its context.
        if isinstance(self._source, self.__class__):
            return self._source.get_resource(required_type)
        required_type = _get_required_type(required_type)
        if required_type not in self._cache:
            out = get_handle_from_resource(self._source, required_type)
            self._cache[required_type] = out
        return self._cache[required_type]

    def __enter__(self):
        """Entering context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Simply ensure all file handles are closed."""
        for handle in self._cache.values():
            getattr(handle, "close", lambda: None)()



"""Tests for IO utilities."""
from __future__ import annotations

from io import BufferedReader, BufferedWriter
from pathlib import Path

import pytest
from tables import File

from dascore.utils.io import (
    BinaryReader,
    BinaryWriter,
    HDF5Reader,
    HDF5Writer,
    IOResourceManager,
    get_handle_from_resource,
)


class _BadType:
    """A dummy type for testing."""


def _dummy_func(arg: str, arg2: _BadType) -> int:
    """A dummy function."""


class TestGetHandleFromResource:
    """Tests for getting the file handle from specific resources."""

    def test_bad_type(self):
        """
        In order to not break anything, unsupported types should just
        return the original argument.
        """
        out = get_handle_from_resource("here", _BadType)
        assert out == "here"

    def test_path_to_buffered_reader(self, tmp_path):
        """Ensure we get a reader from tmp path reader."""
        path = tmp_path / "test_read_buffer.txt"
        path.touch()
        handle = get_handle_from_resource(path, BinaryReader)
        assert isinstance(handle, BufferedReader)
        handle.close()

    def test_path_to_buffered_writer(self, tmp_path):
        """Ensure we get a reader from tmp path reader."""
        path = tmp_path / "test_buffered_writer.txt"
        handle = get_handle_from_resource(path, BinaryWriter)
        assert isinstance(handle, BufferedWriter)
        handle.close()

    def test_path_to_hdf5_reader(self, generic_hdf5):
        """Ensure we get a reader from tmp path reader."""
        handle = get_handle_from_resource(generic_hdf5, HDF5Reader)
        assert isinstance(handle, File)
        handle.close()

    def test_path_to_hdf5_writer(self, tmp_path):
        """Ensure we get a reader from tmp path reader."""
        path = tmp_path / "test_hdf_writer.h5"
        handle = get_handle_from_resource(path, HDF5Writer)
        assert isinstance(handle, File)

    def test_get_path(self, tmp_path):
        """Ensure we can get a path."""
        path = get_handle_from_resource(tmp_path, Path)
        assert isinstance(path, Path)

    def test_get_str(self, tmp_path):
        """Ensure we can get a string."""
        my_str = get_handle_from_resource(tmp_path, str)
        assert isinstance(my_str, str)

    def test_already_file_handle(self, tmp_path):
        """Ensure an input that is already the requested type works."""
        path = tmp_path / "pass_back.txt"
        with open(path, "wb") as fi:
            out = get_handle_from_resource(fi, BinaryWriter)
            assert out is fi

    def test_not_implemented(self):
        """Tests for raising not implemented errors for types not supported."""
        bad_instance = _BadType()
        with pytest.raises(NotImplementedError):
            get_handle_from_resource(bad_instance, BinaryReader)
        with pytest.raises(NotImplementedError):
            get_handle_from_resource(bad_instance, BinaryWriter)
        with pytest.raises(NotImplementedError):
            get_handle_from_resource(bad_instance, HDF5Writer)
        with pytest.raises(NotImplementedError):
            get_handle_from_resource(bad_instance, HDF5Reader)


class TestIOResourceManager:
    """Tests for the IO resource manager."""

    def test_basic_context_manager(self, tmp_path):
        """Ensure it works as a context manager."""
        write_path = tmp_path / "io_writer"

        with IOResourceManager(write_path) as man:
            my_str = man.get_resource(_dummy_func)
            assert isinstance(my_str, str)
            path = man.get_resource(Path)
            assert isinstance(path, Path)
            hf = man.get_resource(HDF5Writer)
            fi = man.get_resource(BinaryWriter)
            # Why didn't pytables implement the stream like pythons?
            assert hf.isopen
            assert not fi.closed
        # after the context manager exists everything should be closed.
        assert not hf.isopen
        assert fi.closed

    def test_nested_context(self, tmp_path):
        """Ensure nested context works as well."""
        write_path = tmp_path / "io_writer"
        with IOResourceManager(write_path) as man:
            fi1 = man.get_resource(BinaryWriter)
            with IOResourceManager(man):
                fi2 = man.get_resource(BinaryWriter)
                # nested IOManager should just return value from previous
                assert fi1 is fi2
            # on first exist the resource should remain open
            assert not fi2.closed
        # then closed.
        assert fi2.closed

    def test_closed_after_exception(self, tmp_path):
        """Ensure the file resources are closed after an exception."""
        path = tmp_path / "closed_resource_test.txt"
        path.touch()
        try:
            with IOResourceManager(path) as man:
                fi = man.get_resource(BinaryReader)
                raise ValueError("Waaagh!")
        except ValueError:
            assert fi.closed



"""
A module for applying transformation to Patches.

Transforms are defined as
"""
from __future__ import annotations

from dascore.utils.misc import MethodNameSpace

from .differentiate import differentiate
from .fft import rfft
from .fourier import dft, idft
from .integrate import integrate
from .spectro import spectrogram
from .strain import velocity_to_strain_rate


class TransformPatchNameSpace(MethodNameSpace):
    """Patch namesapce for transformations."""

    velocity_to_strain_rate = velocity_to_strain_rate
    spectrogram = spectrogram

    rfft = rfft
    dft = dft
    idft = idft

    differentiate = differentiate
    integrate = integrate



"""
Modules for reading and writing fiber data.
"""
from __future__ import annotations
from dascore.io.core import FiberIO, read, scan, scan_to_df, write, PatchFileSummary
from dascore.utils.io import BinaryReader, BinaryWriter, HDF5Reader, HDF5Writer
from dascore.utils.misc import MethodNameSpace


class PatchIO(MethodNameSpace):
    write = write



"""A 2D trace object."""
from __future__ import annotations

import warnings
from collections.abc import Mapping, Sequence

import numpy as np
from rich.text import Text

import dascore as dc
import dascore.proc.coords
from dascore.compat import DataArray, array
from dascore.core.attrs import PatchAttrs
from dascore.core.coordmanager import CoordManager, get_coord_manager
from dascore.core.coords import BaseCoord
from dascore.io import PatchIO
from dascore.transform import TransformPatchNameSpace
from dascore.utils.display import array_to_text, attrs_to_text, get_dascore_text
from dascore.utils.models import ArrayLike
from dascore.viz import VizPatchNameSpace


class Patch:
    """
    A Class for managing data and metadata.

    Parameters
    ----------
    data
        The array data representing fiber optic measurements.
    coords
        The coordinates, or dimensional labels for the data.
        A few types of input are permitted. If a mapping (eg dict) the value
        should conform to one of the following three forms:
        {coord_name: coord}
        {coord_name: ((dimensions,), coord)}
        {coord_name: (dimensions, coord)}
        Where coord can be a numpy array or a
        [`BaseCoord`](`dascore.core.coords.BaseCoord`) object.
        A [`CoordManager`](`dascore.core.coordmanager.CoordManager`) is also acceptable.
    dims
        A sequence of dimension strings. The first entry corresponds to the
        first axis of data, the second to the second dimension, and so on.
    attrs
        Optional attributes (non-coordinate metadata) passed as a dict or
        [PatchAttrs](`dascore.core.attrs.PatchAttrs')

    Notes
    -----
    - If coordinates and dims are not provided, they will be extracted from
    attrs, if possible.

    - If coords and attrs are provided, attrs will have priority. This means
    if there is a conflict between information contained in both, the coords
    will be recalculated.
    """

    data: ArrayLike
    coords: CoordManager
    dims: tuple[str, ...]
    attrs: PatchAttrs | Mapping

    def __init__(
        self,
        data: ArrayLike | DataArray | None = None,
        coords: Mapping[str, ArrayLike | BaseCoord] | CoordManager | None = None,
        dims: Sequence[str] | None = None,
        attrs: Mapping | PatchAttrs | None = None,
    ):
        if isinstance(data, DataArray | self.__class__):
            data, attrs, coords = data.data, data.attrs, data.coords
        if dims is None and isinstance(coords, CoordManager):
            dims = coords.dims
        # Try to generate coords from ranges in attrs
        if coords is None and attrs is not None:
            attrs = dc.PatchAttrs.from_dict(attrs)
            coords = attrs.coords_from_dims()
            dims = dims if dims is not None else attrs.dim_tuple
        # Ensure required info is here
        non_attrs = [x is None for x in [data, coords, dims]]
        if any(non_attrs) and not all(non_attrs):
            msg = "data, coords, and dims must be defined to init Patch."
            raise ValueError(msg)
        coords = get_coord_manager(coords, dims=dims)
        # the only case we allow attrs to include coords is if they are both
        # dicts, in which case attrs might have unit info for coords.
        if isinstance(attrs, Mapping) and attrs:
            coords, attrs = coords.update_from_attrs(attrs)
        else:
            # ensure attrs conforms to coords
            attrs = dc.PatchAttrs.from_dict(attrs).update(coords=coords)
        self._coords = coords
        self._attrs = attrs
        self._data = array(self.coords.validate_data(data))

    def __eq__(self, other):
        """Compare one Patch."""
        return dascore.proc.equals(self, other)

    def __add__(self, other):
        return dascore.proc.apply_operator(self, other, np.add)

    def __sub__(self, other):
        return dascore.proc.apply_operator(self, other, np.subtract)

    def __floordiv__(self, other):
        return dascore.proc.apply_operator(self, other, np.floor_divide)

    def __truediv__(self, other):
        return dascore.proc.apply_operator(self, other, np.divide)

    def __mul__(self, other):
        return dascore.proc.apply_operator(self, other, np.multiply)

    def __pow__(self, other):
        return dascore.proc.apply_operator(self, other, np.power)

    def __rich__(self):
        dascore_text = get_dascore_text()
        patch_text = Text("Patch ⚡", style="bold")
        header = Text.assemble(dascore_text, " ", patch_text)
        line = Text("-" * len(header))
        coords = self.coords.__rich__()
        attrs = self.attrs
        data = array_to_text(self.data, units=attrs.get("data_units"))
        attrs = attrs_to_text(self.attrs)
        out = Text("\n").join([header, line, coords, data, attrs])
        return out

        pass

    def __str__(self):
        out = self.__rich__()
        return str(out)

    __repr__ = __str__

    @property
    def dims(self) -> tuple[str, ...]:
        """Return the dimensions contained in patch."""
        return self.coords.dims

    @property
    def coord_shapes(self) -> dict[str, tuple[int, ...]]:
        """Return a dict of coordinate: (shape, ...)."""
        return self.coords.coord_shapes

    @property
    def attrs(self) -> PatchAttrs:
        """Return the dimensions contained in patch."""
        return self._attrs

    @property
    def coords(self) -> CoordManager:
        """Return the dimensions contained in patch."""
        return self._coords

    @property
    def data(self) -> ArrayLike:
        """Return the dimensions contained in patch."""
        return self._data

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the data array."""
        return self.coords.shape

    @property
    def size(self) -> tuple[int, ...]:
        """Return the shape of the data array."""
        return self.coords.size

    # --- basic patch functionality.

    new = dascore.proc.new
    equals = dascore.proc.equals
    update_attrs = dascore.proc.update_attrs
    assert_has_coords = dascore.proc.assert_has_coords
    get_coord = dascore.proc.get_coord
    to_xarray = dascore.proc.to_xarray
    pipe = dascore.proc.pipe
    set_dims = dascore.proc.set_dims
    squeeze = dascore.proc.squeeze
    transpose = dascore.proc.transpose
    snap_coords = dascore.proc.snap_coords
    sort_coords = dascore.proc.sort_coords
    rename_coords = dascore.proc.rename_coords
    update_coords = dascore.proc.update_coords

    def assign_coords(self, *args, **kwargs):
        """Deprecated method for update_coords."""
        msg = "assign_coords is deprecated, use update_coords instead."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.update_coords(*args, **kwargs)

    set_units = dascore.proc.set_units
    convert_units = dascore.proc.convert_units
    simplify_units = dascore.proc.simplify_units

    # --- processing funcs

    select = dascore.proc.select
    iselect = dascore.proc.iselect
    decimate = dascore.proc.decimate
    detrend = dascore.proc.detrend
    dropna = dascore.proc.dropna
    pass_filter = dascore.proc.pass_filter
    sobel_filter = dascore.proc.sobel_filter
    median_filter = dascore.proc.median_filter
    aggregate = dascore.proc.aggregate
    abs = dascore.proc.abs
    real = dascore.proc.real
    imag = dascore.proc.imag
    angle = dascore.proc.angle
    resample = dascore.proc.resample
    iresample = dascore.proc.iresample
    interpolate = dascore.proc.interpolate
    normalize = dascore.proc.normalize
    standardize = dascore.proc.standardize
    taper = dascore.proc.taper
    rolling = dascore.proc.rolling

    # --- Method Namespaces
    # Note: these can't be cached_property (from functools) or references
    # to self stick around and keep large arrays in memory.

    @property
    def viz(self) -> VizPatchNameSpace:
        """The visualization namespace."""
        return VizPatchNameSpace(self)

    @property
    def tran(self) -> TransformPatchNameSpace:
        """The transformation namespace."""
        return TransformPatchNameSpace(self)

    @property
    def io(self) -> PatchIO:
        """Return a patch IO object for saving patches to various formats."""
        return PatchIO(self)



"""Tests for coordinate processing methods."""
from __future__ import annotations

import numpy as np


class TestSortCoords:
    """Test sorting patches' coordinates."""

    def test_forward_sort(self, wacky_dim_patch):
        """Test sort both dims forward."""
        # default sort should just sort both dims.
        out = wacky_dim_patch.sort_coords()
        for name, coord in out.coords.coord_map.items():
            if name not in out.dims:
                continue
            assert coord.sorted

    def test_reverse_sort(self, wacky_dim_patch):
        """Test sort both dims backward."""
        out = wacky_dim_patch.sort_coords(reverse=True)
        for name, coord in out.coords.coord_map.items():
            if name not in out.dims:
                continue
            assert coord.reverse_sorted

    def test_data_sorted_correctly(self, wacky_dim_patch):
        """Simple test to ensure data were sorted correctly."""
        patch = wacky_dim_patch
        shape = patch.shape
        dims = patch.dims
        # get array that counts cols and rows
        array_list = np.meshgrid(
            np.arange(shape[0]), np.arange(shape[1]), indexing="ij"
        )
        assert all([x.shape == patch.shape for x in array_list])
        # iterate each array and manually test sorting of data matches expected.
        for dim, array, ind in zip(dims, array_list, (1, 0)):
            coord = patch.coords.coord_map[dim]
            arg_sort = np.argsort(coord.values)
            new = patch.new(data=array).sort_coords(dim)
            data_along_slice = np.take(new.data, 0, ind)
            assert np.all(np.equal(arg_sort, data_along_slice))


class TestSnapDims:
    """Tests for snapping dimensions."""

    def test_snap_monotonic(self, wacky_dim_patch):
        """Ensure we can snap a single monotonic coordinate."""
        out = wacky_dim_patch.snap_coords("time")
        coord = out.coords.coord_map["time"]
        assert coord.sorted
        assert coord.evenly_sampled

    def test_snap_array(self, wacky_dim_patch):
        """Ensure we can snap a non monotonic coordinate."""
        out = wacky_dim_patch.snap_coords("distance")
        coord = out.coords.coord_map["distance"]
        assert coord.sorted
        assert coord.evenly_sampled

    def test_snap_dims(self, wacky_dim_patch):
        """Ensure we can snap a non monotonic coordinate."""
        out = wacky_dim_patch.snap_coords()
        for dim in out.dims:
            coord = out.coords.coord_map[dim]
            assert coord.sorted
            assert coord.evenly_sampled
